(** Importation of an EVA specification in abstract syntax (multi-process)
  into a list of clauses *)

open Error;;
open Trace;;
open Location;;
open Termutil;;
open Abstract;;
open Types;;
open Translator;; (* for symbol list construction *)
open Symbols;;
open Pp;;

open Clause;;
open Signature;;

(***************************************************************
 * instructions roles and labels 
 ***************************************************************)

(** [get_first_label il]
  @return the label of the first non [skip]
  instruction of the instructions list [il] 
  if it has a label, or ["begin"] if it has no label,
  or [""] (empty string if [il] is empty or contains only
  SKIPs. *)
let rec get_first_label = function 
    [] -> 
      (warning "role without instructions");
      (* no instructions *)
      "";
  | i::l -> (
      match i.lin_instr with
          (* skip the SKIP! *)
          EVA_SKIP -> 
            get_first_label l
        | EVA_MSG (l, _, _, _) -> 
            (string_of_llabel l)
          (* other instructions have no label, invent one *)
        | _ -> 
            "begin");;

let incr_label label =
  label^"_next";;

let last_label () = 
  "end.";;

let next_label label i = 
  match i.lin_instr with
      EVA_MSG  (l, _, _, _) -> 
        string_of_llabel l
    | EVA_SKIP ->
        raise (Internal_error "next_label: unexpected SKIP")
    | EVA_ASSIGN (_, _, _)
    | EVA_COMP (_, _, _) -> 
        incr_label label
    | EVA_BLOCK _ 
    | EVA_SWITCH (_, _)  ->
        raise (Internal_error "next_label");;
     
(* roles_lprotocol pr
 * returns a list of pairs ((string)role, first label)
 * from the given located protocol pr *)
let rec roles_lprotocol pr = 
  match pr.lpr_protocol with
      (EVA_MP l) -> 
        raise (Internal_error "roles_lprotocol: spec a la BAN")
    | (EVA_PROCESS pl) -> 
        (List.map get_role pl)
and get_role { pr_role = r; pr_messages = il } =
  ((string_of_lid r), (get_first_label il));;
  


(***************************************************************
 * Horn clauses of knowledge declarations 
 * (principals initial states)
 ***************************************************************)

(** [state_ldcllist d (r, l)]
  @return the initial state built from the 
  given located KNOWS declaration [d]
  found in the given list [dl] of EVA located declarations
  associated to role [r], and with label [l]. *)
let rec state_ldcllist kd (r, l) =
  (* the args of the initial states are
     the delocated terms declared with KNOWS in [kd]. *)
  let args = 
    try 
      (match kd.ldc_dcl with
           EVA_DCL_KNOW (_, kl) ->
             (List.map (fun x -> x.lte_term) kl)
         | _ -> raise (Internal_error "state_ldcllist"))
    with 
        Not_found -> []
  in { st_role = r; 
       st_label = l; 
       st_args = args }
(** [get_know_ldcl r dl]
  @return the first located declaration of the form [KNOWS]
  associated to the (string) role [r] in the given list [dl]
  of EVA located declarations.
  @raise Not_found if [dl] contains no declaration with role [r]. *)
and get_know_ldcl r = function
    [] -> raise Not_found
  | d::l -> (
      match d.ldc_dcl with
          EVA_DCL_KNOW (i, kl) ->
            if ((string_of_lid i) = r)
            then d
            else (get_know_ldcl r l)
        | _ -> (get_know_ldcl r l) );;


(** [initstates_of_roles dl rl]
  - param [dl] list of located declarations
  - param [rl]  list of pairs [(r,l)] of role and label,
  @return the list of initial states associated
  to the elements of [rl],
  each built from the first located KNOWS declaration
  found in the given list [dl] 
  associated to role [r]. *)
let rec initstates_of_roles dl = function
    [] -> []
  | (r, l)::rll -> 
      (* there are no instruction, see {!get_first_label} *)
      if (l == "") 
      then 
        initstates_of_roles dl rll
      else 
        let d = (get_know_ldcl r dl) 
        in (state_ldcllist d (r, l))::(initstates_of_roles dl rll);;
    

(***************************************************************
 * Horn clauses of symbols declarations 
 * (interpretation domain)
 ***************************************************************)

let rec string_of_type = function
    EVA_VOID -> "void"
  | EVA_PRINCIPAL -> "p"
  | EVA_NUMBER -> "n"
  | EVA_AALGO -> "aa"
  | EVA_SALGO -> "sa"
  | EVA_TALGO -> "ta"
  | EVA_USERTYPE s 
  | EVA_BASETYPE s -> s
  | EVA_FUN t -> (string_of_type t)^"*number"
  | EVA_HFUN t -> (string_of_type t)^"*hnumber"
  | EVA_KEY t -> (string_of_type t)^"*kp"
;;

(* constant symbols for generic domain for types *)
let constant_of_type t = 
  (string_of_type t)^"0";;

(* remove repetitions *)
let rec clean_list = function
    [] -> []
  | a::l -> 
      if (List.exists (fun x -> (x = a)) l)
      then (clean_list l)
      else a::(clean_list l)
and clean_typelist = function
    [] -> []
  | a::l -> 
      if (List.exists (fun x -> (x = a)) l)
        (* if (isin a l)  *)
      then (clean_typelist l)
      else a::(clean_typelist l)


(***************************************************************
 * Horn clauses of protocol messages
 ***************************************************************)

(* [horn_linstr st lab i]
 * @returns a list of horn clauses computed with:
 * - the state [st] (current state),
 * - the string [lab] (next label),
 * - the located instruction [i] *)
let rec horn_instr st new_label = function
    EVA_SKIP -> []
  | EVA_MSG (_, is, ir, t) -> 
      (* send *)
      if (st.st_role = (string_of_lid is))
      then 
        begin
          let new_state = { st with st_label = new_label } in
            (trace "send: %s(%s) => %s(%s),  %s(%s) => kn\n" 
               st.st_role st.st_label 
               new_state.st_role new_state.st_label
               new_state.st_role new_state.st_label);            
            let clause1 = { cl_neg = [ STATE st ]; 
                            cl_pos = [ STATE new_state ] }
            and clause2 = { cl_neg = [ STATE new_state ]; 
                            cl_pos = [ INTRUK t.lte_term ] }
            in [clause1; clause2]
        end
      (* receive *)
      else if (st.st_role = (string_of_lid ir))
      then
        let args = st.st_args in 
        let new_args = (newargs_lterm args t) in 
        let new_state = { st with st_label = new_label; st_args = args@new_args } in
          (trace "recv: %s%s, kn => %s%s\n" 
             st.st_role st.st_label 
             new_state.st_role new_state.st_label);          
          let clause1 = { cl_neg = [ STATE st; INTRUK t.lte_term ];
                          cl_pos = [ STATE new_state ] }
          in [clause1]
      else raise (Internal_error "state_linstr: msg is not send or recv")
  | EVA_ASSIGN (ir, iv, t) -> 
      if (st.st_role = (string_of_lid ir))
      then
        let args = st.st_args in 
        let new_state = { st with 
                            st_label = new_label; 
                            st_args = (replace_arg iv t.lte_term args) } in
        let clause1 = { cl_neg = [STATE st]; cl_pos = [STATE new_state] } 
        in [clause1]
      else raise (Internal_error "state_linstr: assign")
  | EVA_COMP (ir, iv, t) ->
      if (st.st_role = (string_of_lid ir))
      then        
        let args = st.st_args in 
        let new_args =  (replace_arg iv t.lte_term args)  in
        let new_state = { st with st_label = new_label; st_args = new_args } in
        let clause1 =  { cl_neg = [STATE { st with st_args = new_args } ];
                         cl_pos = [STATE new_state] }
        in [clause1]
      else raise (Internal_error "state_linstr: comp")
  | EVA_BLOCK _ -> 
      raise (Internal_error "horn_instr: block")
  | EVA_SWITCH (t, _) -> 
      raise (Clause_error ("SWITCH is not supported", t.lte_loc))
and horn_linstr st new_label { lin_instr = i } =
  horn_instr st new_label i;
and replace_arg i t = function
    [] -> []
  | (EVA_TERM_ID s)::l -> 
      if (s = (string_of_lid i))
      then t::l
      else (EVA_TERM_ID s)::(replace_arg i t l)
  | l -> l
(* newargs_lterm args t
 * returns the list of EVA_TERM_ID in 
 * the given located term t
 * which are not in the given term list args. *)
and newargs_lterm args lterm = 
  let term = lterm.lte_term 
  in match term with
      EVA_TERM_ID i 
    | EVA_TERM_CST i -> 
        if (is_arg term args)
        then []
        else [term]
    | EVA_P t
    | EVA_A t
    | EVA_SA t
    | EVA_AA t
    | EVA_U (_, t) ->
        newargs_lterm args t
    | EVA_CONS tl ->
        (dflatten (List.map (newargs_lterm args) tl))
    | EVA_APP (i, tl) ->
        (* the function symbol i is not added to the state's knowledge *)
        (dflatten (List.map (newargs_lterm args) tl))
    | EVA_CRYPT (a, k, t) 
    | EVA_SIGN (a, k, t) ->
        let lta = (
          match a with 
              EVA_VANILLA -> []
            | EVA_ALGO ta -> [ta] ) in
        let tl = lta@[k; t] 
        in (dflatten (List.map (newargs_lterm args) tl))
    | _ -> raise (Internal_error "newargs_lterm")
and is_arg t = function
    [] -> false
  | a::l -> 
      if (a = t)
      then true
      else is_arg t l
(* complement l1 l2 = l2 / l1 *)
and complement l1 = function
    [] -> []
  | a::l -> 
      if (is_arg a l1)
      then complement l1 l
      else a::(complement l1 l)
(* disjoint append *)
and dappend l1 = function
    [] -> l1
  | a::l -> 
      if is_arg a l1
      then dappend l1 l
      else a::(dappend l1 l)
(* disjoint flatten *)
and dflatten = function
    [] -> []
  | l::ll -> l@(complement l (dflatten ll));;
    

(** [is_skip i]
  @return whether the given instruction [i]
  is SKIP or not *)
let rec is_skip = function 
    EVA_SKIP -> true 
  | _ -> false
and is_lskip i = 
  is_skip i.lin_instr;;

(* [horn_linstrlist st il]
 * @return a list of transition clauses
 * starting at state [st] and following the
 * located instruction list [il] *)
let rec horn_linstrlist st = function
    [] -> []
  | i::[] -> 
      if (is_lskip i)
      then []
      else horn_linstr st (last_label ()) i
  | i1::(i2::ill) ->
      try   
        (* skip the SKIP *)
        if (is_lskip i1)
        then horn_linstrlist st (i2::ill)
          (* skip the second SKIP *)          
        else if (is_lskip i2)
        then horn_linstrlist st (i1::ill)
        else 
          let clauses = horn_linstr st (next_label st.st_label i2) i1 in
          let st1 = (
            match clauses with
                [] -> st
              | _ -> (first_state_clauselist clauses))
          in clauses@(horn_linstrlist st1 (i2::ill))
      with 
          Not_found -> 
            raise (Internal_error 
                     "horn_linstrlist: first_state_clauselist Not_found")

            
(* horn_process init_state pr
 * returns the clause list
 * built from the given initial state init_state
 * and the given program pr.pr_messages *)
let horn_process init_state { pr_messages = il } = 
  horn_linstrlist init_state il;;

(* horn_protocol sl pr
 * returns the clause list
 * built from the given protocol
 * and the given list of initial states sl. *)
let rec horn_protocol sl = function
    (EVA_MP l) -> 
      raise (Internal_error "horn_protocol: spec a la BAN")
  | (EVA_PROCESS pl) -> 
      (List.flatten (List.map (horn_process1 sl) (unblock pl)))
and horn_lprotocol sl p =
  horn_protocol sl p.lpr_protocol
(* horn_process1 sl pr
 * finds the first state corresp. to pr.role
 * in state list sl 
 * and calls horn_process with this arg and pr. *)
and horn_process1 sl pr = 
  match sl with
      [] -> raise (Internal_error "horn_process1: missing init state")
    | s::l -> 
        let { st_role = r } = s in
          if (r = (string_of_lid pr.pr_role))
          then horn_process s pr
          else  horn_process1 l pr;;


(***************************************************************
 * Horn clauses of sessions
 ***************************************************************)

(* [horn_session stl sl s]
 * returns a list of clauses defining the
 * an initial state according to 
 * the given located session declaration [s]
 * and the given list of initial states [stl],
 * and the given EVA symbol list [sl] (for horn_bang) *)
let rec horn_lsession stl sl s = 
  match s.lse_session with
      EVA_BANG ->
        horn_bang stl sl
    | EVA_CONSTRAINED_SESSIONS cl ->
        (warning "%s Warning:\n constrained session not supported\n" 
           (string_of_location s.lse_loc));
        []
    | EVA_SESSIONS al
    | EVA_SESSION (_, al) ->
        let subst = List.map (fun (i, t) -> ((string_of_lid i), t.lte_term)) al                
        in (List.map 
              (* make pos clauses from states *)
              (fun st -> { cl_neg = []; cl_pos = [STATE st] })
              (List.concat (List.map (instanciate_state s.lse_loc subst) stl)))
and instanciate_state loc subst state =
  if (is_grounding_state subst state)
  then [ (subst_state subst state) ]
  else 
    begin
      (warning 
         "%s Warning:\n this session is not grounding \
          for the initial state of %s\n"
         (string_of_location loc)
         state.st_role); 
      []
    end
and horn_bang stl sl = 
  horny_bang stl sl [] []
(** [horny_bang stl sl tl cl]
  - param [stl]  (initial) state list
  - param [sl] symbol list
  - param [tl]  type list (record types of vars in sl)
  - param [cl]  clause list (clauses already built) 
  @return   *)
and horny_bang stl sl tl cl = 
  match stl with 
      [] -> cl (* (List.concat (List.map horn_generic_type (clean_typelist tl)))@cl *)
    | st::l -> 
        (* list of variables occuring in the terms of the state st *)
        let vars = clean_list (List.concat (List.map term_vars st.st_args)) in 
        (* list of resp. types of the above variables *)
        let types = List.map (type_of_id sl) vars in 
        (* list of type guards (literals) corresp. to the above types *)
        let negs = List.map2 (fun i t -> TYPE (t, (EVA_TERM_ID i))) vars types in
        let claus = { cl_neg = negs; cl_pos = [ STATE st ] }
        in horny_bang l sl (types@tl) (claus::cl)
and type_of_id sl i = 
  try
    let sy = (assoc_symbol i sl) 
    in (get_type sy)
  with
      Symbols_not_found -> raise (Internal_error ("type_of_id: "^i^" not found "))
    | Symbols_error msg -> raise (Internal_error ("type_of_id: "^msg));;
      


(***************************************************************
 * Horn clauses of assume formulas
 ***************************************************************)

let rec supported_assume ql al a =
  (supported_assume_body 0 al)&&(supported_assume_head a)
(** [supported_assume_body c al]
  - param [c] a counter of the number of secret predicates already parsed
  - param [al] a list of atoms
  @return [true] iff the list of atoms [al] is supported *)
and supported_assume_body c = function
    [] -> true
  | a::l -> (
      match a.lat_atom with 
          EVA_FALSE ->
            (warning 
               "%s Warning:\n FALSE in body, assume ignored\n"
               (string_of_location a.lat_loc));
            false
        | EVA_ATOM_EQ _ ->
            (warning 
               "%s Warning:\n equality literals not supported, assume ignored\n"
               (string_of_location a.lat_loc));
            false
        | EVA_ATOM_NEQ _ ->
            (warning 
               "%s Warning:\n disequality literals not supported, assume ignored\n"
               (string_of_location a.lat_loc));
            false    
        | EVA_ATOM_SECRET t ->
            if (c == 0) 
            then 
              if (supported_lterm_assume t)
              then (supported_assume_body 1 l)
              else false
            else 
              begin
                (warning 
                   "%s Warning:\n more than 1 secret literal in clause body, assume ignored\n"
                   (string_of_location a.lat_loc));
                false    
              end
        | EVA_ATOM_HONEST t ->
            if (supported_lterm_assume t)
            then supported_assume_body c l 
            else false
          (* user predicates always supported *)
        | _ -> supported_assume_body c l)
and supported_assume_head a =
  match a.lat_atom with 
      EVA_TRUE ->
        (warning 
           "%s Warning:\n head = TRUE, assume ignored\n"
           (string_of_location a.lat_loc));
        false
    | EVA_ATOM_EQ _ ->
        (warning 
           "%s Warning:\n equality literals not supported, assume ignored\n"
           (string_of_location a.lat_loc));
        false
    | EVA_ATOM_NEQ _ ->
        (warning 
           "%s Warning:\n disequality literals not supported, assume ignored\n"
           (string_of_location a.lat_loc));
        false    
    | EVA_ATOM_SECRET t
    | EVA_ATOM_HONEST t ->
        (supported_lterm_assume t)
      (* user predicates always supported *)
    |  _ -> true
and supported_lterm_assume lterm =
  match lterm.lte_term with
      EVA_TERM_ID _ -> 
        true
    | EVA_TERM_CST _ -> 
        true  
    | EVA_TERM_VALUE _ ->
        true
    | EVA_TERM_ALIAS (a, t) ->
        (supported_lterm_assume t)
    | EVA_LOCATED (i1, i2) ->
        raise (Clause_error ("located variable in assume", lterm.lte_loc))
    | EVA_QUANTIFIED (_, _) -> 
        true
    | EVA_P t
    | EVA_A t
    | EVA_SA t
    | EVA_AA t
    | EVA_U (_, t) ->
        supported_lterm_assume t
    | EVA_CONS tl 
    | EVA_APP (_, tl) ->
        List.fold_left (fun x y -> x&&y) true 
        (List.map supported_lterm_assume tl)
    | EVA_CRYPT (a, k, t) 
    | EVA_SIGN (a, k, t) ->
        (supported_algo_assume a) &&
        (supported_lterm_assume k) &&
        (supported_lterm_assume t)
    | EVA_PCENT (t1, t2) ->
        (supported_lterm_assume t1) && (supported_lterm_assume t2)      
and supported_algo_assume = function
    EVA_VANILLA -> true
  | EVA_ALGO t -> supported_lterm_assume t;;

(* [literal_quantification ql]
 * returns the list of literals corresponding to
 * the list of typed quantified variables ql *)
let rec literal_quantification = function
    [] -> []
  | (i,lt)::l -> 
      let t = lt.lty_type 
      in (TYPE (t, (EVA_QUANTIFIED (string_of_lid i, t))))::
           (literal_quantification l);;

(* horn_assume_lstatement f
 * returns a list of clauses (singleton or empty)
 * corresponding to the given assume located statement f
 * 
 * If the assume clause is not supported, the empty
 * list is returned and a warning is printed on error output. *)
let rec horn_assume_lstatement f = 
  horn_assume_statement f.lst_statement
and  horn_assume_statement = function
    EVA_ASSUME (ql, al, a) -> horn_assume ql al a
  | _ -> []
and horn_assume ql al a = 
  if (supported_assume ql al a)
  then 
    let qu = literal_quantification ql in
    let cl = { cl_neg = qu; cl_pos = [] } 
    in [ (horn_assume_head (horn_assume_body cl al) a) ]
  else []
(* horn_assume_body cl al
 * [al] must be the body of a supported assume. *)
and horn_assume_body cl = function
    [] -> cl
  | a::l -> (
      match a.lat_atom with 
          EVA_TRUE ->
            (horn_assume_body cl l)
            (* (horn_assume_body { cl with cl_neg = (TRUE::cl.cl_neg) } l) *)
          (* should not happen when passed test supported_assume *)       
        | EVA_FALSE ->
            raise (Internal_error "horn_assume_body: false")
            (* (horn_assume_body { cl with cl_neg = (FALSE::cl.cl_neg) } l) *)
        | EVA_ATOM_HONEST t ->
            (horn_assume_body 
               { cl with cl_neg = (HONEST t.lte_term)::cl.cl_neg } l)
          (* transfred as KNOW in head *)
        | EVA_ATOM_SECRET t ->
            (horn_assume_body 
               { cl with cl_pos = (INTRUK t.lte_term)::cl.cl_pos } l)
        | EVA_ATOM_USER (i, t) ->
            (horn_assume_body 
               { cl with 
                   cl_neg = (USER ((string_of_lid i), t.lte_term))::cl.cl_neg }
               l)
        | _  ->
            raise (Internal_error "horn_assume_body"))
(* horn_assume_head cl a
 * a must be the head of a supported assume. *)
and horn_assume_head cl a =
  match a.lat_atom with 
      EVA_TRUE ->
        raise (Internal_error "horn_assume_head: true")
        (* { cl with cl_pos = TRUE::cl.cl_pos } *)
    | EVA_FALSE ->
        cl        (* { cl with cl_pos = FALSE::cl.cl_pos } *)
    | EVA_ATOM_HONEST t ->
        { cl with cl_pos = (HONEST t.lte_term)::cl.cl_pos }
      (* transfered as KNOW in body *)
    | EVA_ATOM_SECRET t ->
        { cl with cl_neg = (INTRUK t.lte_term)::cl.cl_neg }            
    | EVA_ATOM_USER (i, t)  ->
        { cl with cl_pos = (USER ((string_of_lid i), t.lte_term))::cl.cl_pos }       
    | _  ->
        raise (Internal_error "horn_assume_body");;                  

(***************************************************************
 * Horn clauses of claim formulas
 ***************************************************************)

let rec supported_claim al a =
  (supported_claim_body al)&&(supported_claim_head a)
and supported_claim_body = function
    [] -> true
  | a::l -> (
      match a.lat_atom with
          EVA_FALSE ->
            (warning 
               "%s Warning:\n FALSE in body, claim ignored\n"
               (string_of_location a.lat_loc));
            false
        | EVA_ATOM_EQ _ ->
            (warning 
               "%s Warning:\n equality not supported, claim clause ignored\n"
               (string_of_location a.lat_loc));
            false
        | EVA_ATOM_NEQ _ ->
            (warning 
               "%s Warning:\n equality not supported, claim clause ignored\n"
               (string_of_location a.lat_loc));
            false
         (* user predicates always supported *)
        | _ ->
            supported_claim_body l)
and supported_claim_head a = 
  match a.lat_atom with
      EVA_TRUE ->
        (warning 
           "%s Warning:\n head = TRUE, claim ignored\n"
           (string_of_location a.lat_loc));
        false
    | EVA_ATOM_EQ _ ->
        (warning 
           "%s Warning:\n equality not supported, claim clause ignored\n"
           (string_of_location a.lat_loc));
        false
    | EVA_ATOM_NEQ _ ->
        (warning 
           "%s Warning:\n equality not supported, claim clause ignored\n"
           (string_of_location a.lat_loc));
        false
     (* user predicates always supported *)
    | _ -> true;;
  

(** [claim_roles al a]
  @return the list of roles (string) occuring
  in located variables of the given claim formula [al => a] *)
let rec claim_roles al a = 
  List.concat (List.map latom_roles (a::al))
and latom_roles a = 
  match a.lat_atom with
      EVA_ATOM_EQ (t1, t2)
    | EVA_ATOM_NEQ (t1, t2) ->
        (lterm_roles t1)@(lterm_roles t2)
    | EVA_ATOM_HONEST t 
    | EVA_ATOM_SECRET t 
    | EVA_ATOM_USER (_, t) ->
        lterm_roles t
    | _ -> []
and lterm_roles t =
  match t.lte_term with
      EVA_CONS tl 
    (* the root is a constant symbol, hence can not be located *)
    | EVA_APP (_, tl) ->
        List.concat (List.map lterm_roles tl)
    | EVA_CRYPT (a, k, t) 
    | EVA_SIGN (a, k, t) ->
        (algo_roles a)@(lterm_roles k)@(lterm_roles t)
    | EVA_PCENT (t1, t2) ->
        (lterm_roles t1)@(lterm_roles t2)
    | EVA_TERM_ALIAS (a, t) ->
        (lterm_roles t)
    | EVA_LOCATED (i1, i2) ->
        [string_of_lid i2]
    | _  -> []
and algo_roles = function
    EVA_VANILLA -> []
  | EVA_ALGO t -> (lterm_roles t);;

(** [claim_lparams r al a]
  @return the set of parameters located at the given role (string) [r]
  in the given claim formula [al => a] *)
let rec claim_lparams r al a = 
  List.concat (List.map (latom_lparams r) (a::al))
and latom_lparams r a = 
  match a.lat_atom with
      EVA_ATOM_EQ (t1, t2)
    | EVA_ATOM_NEQ (t1, t2) ->
        (lterm_lparams r t1)@(lterm_lparams r t2)
    | EVA_ATOM_HONEST t 
    | EVA_ATOM_SECRET t 
    | EVA_ATOM_USER (_, t) ->
        lterm_lparams r t
    | _ -> []
and lterm_lparams r t =
  match t.lte_term with
      EVA_CONS tl 
    (* the root is a constant symbol, hence can not be located *)
    | EVA_APP (_, tl) ->
        List.concat (List.map (lterm_lparams r) tl)
    | EVA_CRYPT (a, k, t) 
    | EVA_SIGN (a, k, t) ->
        (algo_lparams r a)@(lterm_lparams r k)@(lterm_lparams r t)
    | EVA_PCENT (t1, t2) ->
        (lterm_lparams r t1)@(lterm_lparams r t2)
    | EVA_TERM_ALIAS (a, t) ->
        (lterm_lparams r t)
    | EVA_LOCATED (i1, i2) ->
        if ((string_of_lid i2) = r)
        then [(string_of_lid i1)]
        else []
    | _  -> 
        []
and algo_lparams r = function
    EVA_VANILLA -> []
  | EVA_ALGO t -> (lterm_lparams r t);;
  

(** [cover_params_states pl sl]
  @returns the sublist of states of the given state list [sl]
  which contain all the parameters of the given list [pl]. *)
let rec cover_params_states pl sl =
  List.concat (List.map (cover_params_state pl) sl)
(** [cover_params_state pl s]
  @return the singleton list [s] if the given state [s]
  contains all the parameters of the given list [pl];
  returns [] otherwise *)
and cover_params_state pl s =
  match pl with
      [] -> [s]
    | p::l -> 
        if (cover_param_state p s)
        then (cover_params_state l s)
        else []
(* cover_param_state p s
 * returns true if the given state s
 * contains the given parameter p *)
and cover_param_state p s = 
   cover_param_terms p s.st_args
and cover_param_terms p = function
    [] -> false
  | t::l -> (
      match t with
          EVA_TERM_ID s ->
            if (s = p)
            then true
            else (cover_param_terms p l)
        | _ ->
            (cover_param_terms p l));;

(*  list_product ll
 * returns the cartesian product of the given 
 * list of list ll *)
let rec list_product = function
    [] -> [[]]
  | l::ll -> 
      let p = (list_product ll)
      in (List.concat (List.map (fun x -> (List.map (fun y -> x::y) p)) l));;

(** [horn_claim_lstatement rl stl f]
  @return the list of clauses associated to the 
  given claim located statement [f]
  given a role list [rl] and a set of states [stl]. *)
let rec horn_claim_lstatement rl stl f = 
  horn_claim_statement rl stl f.lst_statement    
and horn_claim_statement rl stl = function
    EVA_CLAIM (_, al, a) -> 
      if (supported_claim al a)
      then horn_claim stl al a
      else []
  | _ -> []
and horn_claim stl al a = 
  (* the list of roles occurring in located vars of [al=>a] *)
  let rl = (claim_roles al a) in
  (* for each role r in rl, the list of states of stl with role r *)
  let srl = (List.map (fun r -> (states_role r stl)) rl) in
  (* for each role [r] of [rl] the list of parameters located at [r] in of [al a] *)
  let rpl = (List.map (fun r -> (claim_lparams r al a)) rl) in
  (* for each state list of [srl], the sublist of states covering all the
     parameters of the corresponding list in [rpl] *)
  let srl2 = (List.map2 (fun x y -> (cover_params_states y x)) srl rpl) in
  let srl3 = list_product srl2 in
  let srl4 = List.map (fun x -> (List.map (fun y -> (STATE y)) x)) srl3 in
  let c = horn_claim_latoms al a 
  in (List.map (fun x -> { cl_neg = (x@c.cl_neg); cl_pos = c.cl_pos }) srl4)
and horn_claim_latoms al a = 
  let c = 
    match a.lat_atom with
        EVA_TRUE ->
          raise (Internal_error "horn_claim_latoms: true")
      | EVA_FALSE ->
          { cl_neg = []; cl_pos = [] }
      | EVA_ATOM_EQ _ ->
          raise (Internal_error "horn_claim_latoms: eq")
      | EVA_ATOM_NEQ _ ->
          raise (Internal_error "horn_claim_latoms: neq")
      | EVA_ATOM_HONEST t ->
          { cl_neg = []; cl_pos = [(literal_of_latom a)] } 
                               (* [HONEST t.lte_term] *)
      | EVA_ATOM_SECRET t ->
          { cl_neg = [(literal_of_latom a)]; cl_pos = [] }
                  (* [INTRUK t.lte_term] *)
      | EVA_ATOM_USER (i, t) ->
          { cl_neg = []; cl_pos = [(literal_of_latom a)] } 
                               (* [USER ((string_of_lid i), t.lte_term)] *)
  in horn_claim_body c al
and horn_claim_body c = function
    [] -> c
    (* true: ignore *)
  | { lat_atom = EVA_TRUE }::l ->
      horn_claim_body c l
  | a::l ->
      horn_claim_body { c with cl_neg = (literal_of_latom a)::c.cl_neg } l
(* the given atom a must not be an equality or disequality *)
and literal_of_latom a =
  match a.lat_atom with
      (* should not happen when called by horn_claim_body *)
      EVA_TRUE ->
        raise (Internal_error "literal_of_latom: false")       
      (* should not happen when passed test supported_claim *)       
    | EVA_FALSE ->
        raise (Internal_error "literal_of_latom: false")
    | EVA_ATOM_EQ _ ->
        raise (Internal_error "literal_of_latom: eq")
    | EVA_ATOM_NEQ _ ->
        raise (Internal_error "literal_of_latom: neq")
    | EVA_ATOM_HONEST t ->
        HONEST t.lte_term
    | EVA_ATOM_SECRET t ->
        INTRUK t.lte_term
    | EVA_ATOM_USER (i, t) ->
        USER ((string_of_lid i), t.lte_term);;





(***************************************************************
 * clauses of spec
 ***************************************************************)

(* states_of_transitions cl
 * returns the list of states found in the given list of
 * clauses cl, except initial states.
 *
 * NOTE: DIRTY,
 * assume that the clauses of cl are transitions clauses
 * of one of the following forms:
 * - Ai, K => Ai+1  (recv, Ai+1 is returned)
 * - Ai   => Ai+1 ; Ai+1 => K  
 *   (send, in this order, first Ai+1 is returned)
 * - Ai => Ai+1     (assign, compare, Ai+1 is returned) *)
let rec states_of_transitions = function
    [] -> []
    (* receive *)
  | { cl_neg = [ STATE _; INTRUK _ ]; cl_pos = [ STATE s ] }::l ->
      s::(states_of_transitions l)
    (* send *)
  | { cl_neg = [ STATE _ ]; cl_pos = [ STATE s ] }::
      ({ cl_neg = [ STATE _ ]; cl_pos = [ INTRUK _ ] }::l) ->
      s::(states_of_transitions l)
    (* assign, compare *)
  | { cl_neg = [ STATE _ ]; cl_pos = [ STATE s ] }::l ->
      s::(states_of_transitions l)
  | _ -> raise (Internal_error "states_of_transitions: unexpected case");;

(* horn_spec sl sp
 * returns the list of horn clauses associated to 
 * the given spec sp with symbol list sl *)
let rec horn_spec sl sp = 
  let roles =  roles_lprotocol sp.sp_protocol in 
  let init_states = initstates_of_roles sp.sp_declarations roles in
  let domain = Signature.horn_signature sl in
  let sessions = 
    List.concat (List.map (horn_lsession init_states sl) sp.sp_sessions) in 
  let transitions = horn_lprotocol init_states sp.sp_protocol in
  let assumes = List.concat (List.map horn_assume_lstatement 
                               sp.sp_statements) in
  let states = init_states@(states_of_transitions transitions) in
  let roles1 = List.map fst roles in
  let claims = List.concat (List.map (horn_claim_lstatement roles1 states) 
                              sp.sp_statements) 
  in ([ COMMENT "Symbols type declarations" ]@
      (List.map axiomatize domain)@
      [ COMMENT "Initial state of principals" ]@
      (List.map axiomatize sessions)@
      [ COMMENT "Transitions" ]@
      (List.map axiomatize transitions)@
      [ COMMENT "Initial state of intruder" ]@
      (List.map axiomatize assumes)@
      [ COMMENT "Intruders theory" ]@
      (List.map axiomatize (Intruder.rules sl))@
      [ COMMENT "Conjectures" ]@
      (List.map conjecturize claims))
and axiomatize c = 
  AXIOM c
and conjecturize c =
  CONJ c;;




